// The following code runs speed benchmarks for the different modules
// The first inference speed for each module is discarded due to lazy initialization of models

#include <chrono>
#include <iostream>
#include <string>
#include <vector>
#include <cstdlib> 

#include "tf_sdk.h"

using namespace Trueface;

typedef std::chrono::high_resolution_clock Clock;

void benchmarkFaceRecognition(const std::string& license, FacialRecognitionModel model, bool enableGPU, unsigned int batchSize = 1, unsigned int numIterations = 100);
void benchmarkObjectDetection(const std::string& license, bool enableGPU, unsigned int numIterations = 100);
void benchmarkFaceLandmarkDetection(const std::string& license, bool enableGPU, unsigned int numIterations = 100);

int main() {
    // TODO: Replace with your license
    const std::string license = "${LICENSE_CODE}";

    std::cout << "Running speed benchmarks with 1280x720 image\n";

    bool enableGPU = false;
    if (enableGPU) {
        std::cout << "Using GPU for inference" << std::endl;
    }

    benchmarkFaceLandmarkDetection(license, enableGPU);
    if (!enableGPU) {
        // Trueface::SDK::getFaceFeatureVectors is not supported by the LITE and LITE_V2 models.
        benchmarkFaceRecognition(license, FacialRecognitionModel::LITE, enableGPU, 1,  200);
        benchmarkFaceRecognition(license, FacialRecognitionModel::LITE_V2, enableGPU, 1,  200);
    }
    benchmarkFaceRecognition(license, FacialRecognitionModel::TFV5, enableGPU, 1,  40);
    benchmarkFaceRecognition(license, FacialRecognitionModel::FULL, enableGPU, 1,  40);
    benchmarkObjectDetection(license, enableGPU);

    // Benchmarks with batching.
    // On CPU, should be the same speed as a batch size of 1.
    // On GPU, will increase the throughput.
    benchmarkFaceRecognition(license, FacialRecognitionModel::TFV5, enableGPU, 4, 40);

    return 0;
}

std::string getModelName(FacialRecognitionModel model) {
    if (model == FacialRecognitionModel::TFV5) {
        return "TFV5";
    } else if (model == FacialRecognitionModel::FULL) {
        return "FULL";
    } else if (model == FacialRecognitionModel::LITE) {
        return "LITE";
    } else if (model == FacialRecognitionModel::LITE_V2) {
        return "LITE V2";
    } else {
        throw std::runtime_error("The model is currently not supported by the benchmarking script");
    }
}

void benchmarkFaceRecognition(const std::string& license, FacialRecognitionModel model, bool enableGPU, unsigned int batchSize, unsigned int numIterations) {
    // Initialize the SDK
    ConfigurationOptions options;
    options.enableGPU = enableGPU;
    options.frModel = model;

    // Since we initialize the module, we do not need to discard the first inference time.
    InitializeModule initializeModule;
    initializeModule.faceRecognizer = true;
    options.initializeModule = initializeModule;

    SDK tfSdk(options);
    bool valid = tfSdk.setLicense(license);

    if (!valid) {
        std::cout << "Error: the provided license is invalid." << std::endl;
        exit (EXIT_FAILURE);
    }

    // Load the image
    ErrorCode errorCode = tfSdk.setImage("../images/headshot.jpg");
    if (errorCode != ErrorCode::NO_ERROR) {
        std::cout << "Error: could not load the image" << std::endl;
        return;
    }

    // Obtain the aligned chip
    FaceBoxAndLandmarks faceBoxAndLandmarks;
    bool found;
    tfSdk.detectLargestFace(faceBoxAndLandmarks, found);

    uint8_t faceBuffer[112*112*3];
    tfSdk.extractAlignedFace(faceBoxAndLandmarks, faceBuffer);

    std::vector<uint8_t*> alignedFaceImages;
    for (size_t i = 0; i < batchSize; ++i) {
        alignedFaceImages.push_back(faceBuffer);
    }

    std::vector<Faceprint> faceprints;

    // Time the creation of the feature vector
    auto t1 = Clock::now();
    for (size_t i = 0; i < numIterations; ++i) {
        tfSdk.getFaceFeatureVectors(alignedFaceImages, faceprints);
    }
    auto t2 = Clock::now();
    double totalTime = std::chrono::duration_cast<std::chrono::milliseconds>(t2 - t1).count();

    std::cout << "Average time face recognition " << getModelName(model) << ": " << totalTime / numIterations / static_cast<float>(batchSize)
              << " ms | batch size = " << batchSize << " | " << numIterations << " iterations" << std::endl;

}

void benchmarkObjectDetection(const std::string& license, bool enableGPU, unsigned int numIterations) {
    // Initialize the SDK with the fast object detection model
    ConfigurationOptions options;
    options.enableGPU = enableGPU;
    options.objModel = ObjectDetectionModel::FAST;

    // Since we initialize the module, we do not need to discard the first inference time.
    InitializeModule initializeModule;
    initializeModule.faceDetector = true;
    options.initializeModule = initializeModule;

    SDK tfSdk(options);
    bool valid = tfSdk.setLicense(license);

    if (!valid) {
        std::cout << "Error: the provided license is invalid." << std::endl;
        exit (EXIT_FAILURE);
    }

    // Load the image
    ErrorCode errorCode = tfSdk.setImage("../images/bike.jpg");
    if (errorCode != ErrorCode::NO_ERROR) {
        std::cout << "Error: could not load the image" << std::endl;
        return;
    }

    // Discard first inference speed
    std::vector<BoundingBox> boundingBoxes;

    // Time the creation of the feature vector
    auto t1 = Clock::now();
    for (size_t i = 0; i < numIterations; ++i) {
        tfSdk.detectObjects(boundingBoxes);

    }
    auto t2 = Clock::now();
    double totalTime = std::chrono::duration_cast<std::chrono::milliseconds>(t2 - t1).count();

    std::cout << "Average time object detection (fast mode): " << totalTime / numIterations
              << " ms | " << numIterations << " iterations" << std::endl;

}

void benchmarkFaceLandmarkDetection(const std::string& license, bool enableGPU, unsigned int numIterations) {
    // Initialize the SDK
    ConfigurationOptions options;
    options.enableGPU = enableGPU;
    options.smallestFaceHeight = 40;

    // Since we initialize the module, we do not need to discard the first inference time.
    InitializeModule initializeModule;
    initializeModule.faceDetector = true;
    options.initializeModule = initializeModule;

    SDK tfSdk(options);
    bool valid = tfSdk.setLicense(license);

    if (!valid) {
        std::cout << "Error: the provided license is invalid." << std::endl;
        exit (EXIT_FAILURE);
    }

    // Load the image
    ErrorCode errorCode = tfSdk.setImage("../images/headshot.jpg");
    if (errorCode != ErrorCode::NO_ERROR) {
        std::cout << "Error: could not load the image" << std::endl;
        return;
    }

    // Discard first inference speed
    FaceBoxAndLandmarks faceBoxAndLandmarks;
    bool found = false;

    // Time the creation of the feature vector
    auto t1 = Clock::now();
    for (size_t i = 0; i < numIterations; ++i) {
        tfSdk.detectLargestFace(faceBoxAndLandmarks, found);
    }
    auto t2 = Clock::now();
    double totalTime = std::chrono::duration_cast<std::chrono::milliseconds>(t2 - t1).count();

    std::cout << "Average time face and landmark detection: " << totalTime / numIterations
              << " ms | " << numIterations << " iterations" << std::endl;

}